# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.contrib import admin
from django_atlassian.models.connect import SecurityContext

class SecurityContextAdmin(admin.ModelAdmin):
    list_display = ['key']

admin.site.register(SecurityContext, SecurityContextAdmin)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.db.models import fields

class ArrayField(fields.Field):

    def from_jira(self, value, connection):
        return [x.decode(connection.charset) for x in value]

    def get_db_prep_save(self, value, connection):
        if not value:
            return None
        return [x.encode(connection.charset) for x in value]

    def from_db_value(self, value, expression, connection, context):
        """Convert from the database format.

        This should be the inverse of self.get_prep_value()
        """
        return self.to_python(value)

    def to_python(self, value):
        if not value:
            return []
        return value




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.core.exceptions import PermissionDenied

def jwt_required(function):
    def decorator(request, *args, **kwargs):
        sc = getattr(request, 'atlassian_sc', None)
        model = getattr(request, 'atlassian_model', None)
        if not sc or not model:
            raise PermissionDenied
        else:
            return function(request, *args, **kwargs)
    return decorator



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django_atlassian.backends.common import compiler as common_compiler
from django.db.models.sql import compiler

class SQLCompiler(common_compiler.SQLCompiler):
    query_param = 'jql'


class SQLUpdateCompiler(SQLCompiler):
    def as_sql(self):
        # Generate the json to update
        values = {}
        for field, model, val in self.query.values:
            if hasattr(val, 'resolve_expression'):
                val = val.resolve_expression(self.query, allow_joins=False, for_save=True)
                if val.contains_aggregate:
                    raise FieldError("Aggregate functions are not allowed in this query")
            elif hasattr(val, 'prepare_database_save'):
                if field.remote_field:
                    val = field.get_db_prep_save(
                        val.prepare_database_save(field),
                        connection=self.connection,
                    )
                else:
                    raise TypeError(
                        "Tried to update field %s with a model instance, %r. "
                        "Use a value compatible with %s."
                        % (field, val, field.__class__.__name__)
                    )
            else:
                val = field.get_db_prep_save(val, connection=self.connection)
            name = field.column
            if hasattr(val, 'as_sql'):
                sql, params = self.compile(val)
                values[name] = sql
            else:
                values[name] = val 
            
        sql, params = super(SQLUpdateCompiler, self).as_sql()
        params += tuple([{'update_fields': values}])
        return sql, params


    def execute_sql(self, result_type):
        cursor = super(SQLUpdateCompiler, self).execute_sql(result_type)
        try:
            rows = cursor.rowcount if cursor else 0
            is_empty = cursor is None
        finally:
            if cursor:
                cursor.close()
        return rows



# -*- coding: utf-8 -*-

from django.conf import settings
from django.db import connections
from django_atlassian.models.djira import Issue
from django_atlassian.models.confluence import Content

class Router(object):

    def __init__(self):
        self.db_jira_alias = None
        self.db_confluence_alias = None
        for alias, settings_dict in settings.DATABASES.items():
            if settings_dict['ENGINE'] == 'django_atlassian.backends.jira':
                self.db_jira_alias = alias
            if settings_dict['ENGINE'] == 'django_atlassian.backends.confluence':
                self.db_confluence_alias = alias

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        if self.db_confluence_alias == db:
            return False
        if self.db_jira_alias == db:
            return False
        return None

    def db_for_read(self, model, **hints):
        if issubclass(model, Issue):
            # Check if the model has the AtlassianMeta class
            atlassian_meta = getattr(model, 'AtlassianMeta', False)
            if not atlassian_meta:
                return self.db_jira_alias
            atlassian_db = getattr(atlassian_meta, 'db', False)
            if not atlassian_db:
                return self.db_jira_alias
            for alias, settings_dict in connections.databases.items():
                if settings_dict['ENGINE'] == 'django_atlassian.backends.jira' and \
                    alias == atlassian_db:
                    return alias
            return None

        if issubclass(model, Content):
            return self.db_confluence_alias
        return None

    def db_for_write(self, model, **hints):
        if issubclass(model, Issue):
            # Check if the model has the AtlassianMeta class
            atlassian_meta = getattr(model, 'AtlassianMeta', False)
            if not atlassian_meta:
                return self.db_jira_alias
            atlassian_db = getattr(atlassian_meta, 'db', False)
            if not atlassian_db:
                return self.db_jira_alias
            for alias, settings_dict in connections.databases.items():
                if settings_dict['ENGINE'] == 'django_atlassian.backends.jira' and \
                    alias == atlassian_db:
                    return alias

        if issubclass(model, Content):
            return self.db_confluence_alias

        return None



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.db import models

class Content(models.base.Model):
    """
    Base class for all Confluence content models.
    """
    PAGE_TYPE = 'page'
    BLOGPOST_TYPE = 'blogpost'
    COMMENT_TYPE = 'comment'
    ATTACHMENT_TYPE = 'attachment'

    TYPE_CHOICES = (
        (PAGE_TYPE, PAGE_TYPE),
        (BLOGPOST_TYPE, BLOGPOST_TYPE),
        (COMMENT_TYPE, COMMENT_TYPE),
        (ATTACHMENT_TYPE, ATTACHMENT_TYPE),
    )

    id = models.CharField(max_length=255, primary_key=True, unique=True)
    creator = models.CharField(max_length=255, null=False, blank=False)
    created = models.DateField()
    type = models.CharField(max_length=255, choices=TYPE_CHOICES)
    title = models.TextField(null=False, blank=False)
    space = models.CharField(max_length=255, null=False, blank=False)

    def __unicode__(self):
        return str(self.id)

    class Meta:
        abstract = True




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from collections import namedtuple
import requests
import re
import logging
import json

from django.utils.dateparse import parse_datetime
from django.contrib.auth import get_user_model

from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.backends.base.client import BaseDatabaseClient
from django.db.backends.base.creation import BaseDatabaseCreation
from django.db.backends.base.features import BaseDatabaseFeatures
from django.db.backends.base.operations import BaseDatabaseOperations
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.backends.base.validation import BaseDatabaseValidation
from django.db.backends.base.introspection import BaseDatabaseIntrospection, FieldInfo

# Extend the FieldInfo to also have:
# field_name: The model field name
# field_id: The field id
# array_type: In case of being an array, the element's type
FieldInfo = namedtuple('FieldInfo', FieldInfo._fields + ('field_name', 'field_id', 'array_type', 'editable', 'options'))
logger = logging.getLogger('django_atlassian.backends.common')

class AtlassianDatabaseIntrospection(BaseDatabaseIntrospection):
    pass


class AtlassianDatabaseConnection(object):
    def __init__(self, uri, user, password, sc):
        self.uri = uri
        self.user = user
        self.password = password
        self.sc = sc

    def rollback(self):
        return True

    def commit(self):
        return True

    def get_request(self, part):
        uri = self.uri + part
        headers = {'Accept': 'application/json'}
        if self.user and self.password:
            r = requests.get(uri, auth=(self.user, self.password), headers=headers)
        elif self.sc:
            token = self.sc.create_token('GET', uri)
            headers.update({'Authorization': 'JWT {}'.format(token)})
            r = requests.get(uri, headers=headers)
        else:
            return None
        return r

    def put_request(self, part, body):
        json_body = json.dumps(body)
        uri = self.uri + part
        headers = {'Content-Type': 'application/json'}
        if self.user and self.password:
            r = requests.put(uri, data=json_body, auth=(self.user, self.password), headers=headers)
        elif self.sc:
            token = self.sc.create_token('PUT', uri)
            headers.update({'Authorization': 'JWT {}'.format(token)})
            r = requests.put(uri, data=json_body, headers=headers)
        else:
            return None
        return r

    def post_request(self, part, body, header=None, fil=None):
        json_body = json.dumps(body)
        uri = self.uri + part
        headers = {'Content-Type': 'application/json'}
        if header and isinstance(header, dict) and fil:
            headers = header
        if self.user and self.password:
            if fil:
                r = requests.post(uri, files=fil, auth=(self.user, self.password), headers=headers)
            else:
                r = requests.post(uri, data=json_body, auth=(self.user, self.password), headers=headers)
        elif self.sc:
            token = self.sc.create_token('POST', uri)
            headers.update({'Authorization': 'JWT {}'.format(token)})
            if fil:
                r = requests.post(uri, files=fil, headers=headers)
            else:
                r = requests.post(uri, data=json_body, headers=headers)
        else:
            return None
        return r

    def delete_request(self, part):
        uri = self.uri + part
        if self.user and self.password:
            r = requests.delete(uri, auth=(self.user, self.password))
        elif self.sc:
            token = self.sc.create_token('DELETE', uri)
            headers.update({'Authorization': 'JWT {}'.format(token)})
            r = requests.delete(uri)
        else:
            return None
        return r

class AtlassianDatabase(object):
    Error = requests.exceptions.RequestException

    class DatabaseError(Error):
        pass

    class OperationalError(DatabaseError):
        pass

    class IntegrityError(DatabaseError):
        pass

    class DataError(DatabaseError):
        pass

    class InterfaceError(DatabaseError):
        pass

    class InternalError(DatabaseError):
        pass

    class ProgrammingError(DatabaseError):
        pass

    class NotSupportedError(DatabaseError):
        pass


class AtlassianDatabaseValidation(BaseDatabaseValidation):
    pass


class AtlassianDatabaseOperations(BaseDatabaseOperations):
    def quote_name(self, name):
        return name

    def no_limit_value(self):
        return None


class AtlassianDatabaseFeatures(BaseDatabaseFeatures):
    can_use_chunked_reads = False
    supports_transactions = False
    supports_select_union = False
    supports_select_intersection = False
    supports_select_difference = False
    
    def __init__(self, connection):
        self.connection = connection


class AtlassianDatabaseCreation(BaseDatabaseCreation):
    def create_test_db(self, *args, **kwargs):
        """
        Creates a test database, prompting the user for confirmation if the
        database already exists. Returns the name of the test database created.
        """
        pass

    def destroy_test_db(self, *args, **kwargs):
        """
        Destroy a test database, prompting the user for confirmation if the
        database already exists. Returns the name of the test database created.
        """
        pass


class AtlassianDatabaseConvertion(object):
    def __init__(self, connection):
        self.connection = connection

    def extract(self, data, field, raw_field):
        raise NotImplementedError('missing extract implementation')

    def insert(self, data, field, raw_field):
        raise NotImplementedError('missing insert implementation')

    def from_native(self, data, field):
        if field[1] == 'datetime' and data is not None:
            return parse_datetime(data)
        elif field[1] == 'number' and data is not None:
            return float(data)
        elif field[1] in ['user', 'issuetype'] and data is not None:
            return data['name']
        else:
            if field[12] and data.has_key('value'):
                return data['value']
            return data

    def to_native(self, data, field):
        if field[1] == 'datetime' and data is not None:
            return None
        elif field[1] == 'number' and data is not None:
            return float(data)
        elif field[1] in ['user', 'issuetype'] and data is not None:
            return {'name': data}
        else:
            if field[12]:
                return {'value': data}
            return data
        pass


class AtlassianDatabaseCursor(object):
    def __init__(self, connection, db):
        self.connection = connection
        self.db = db
        self.sql = None
        self.fields = None
        self.raw_fields = None
        self.rowcount = 0

    def __normalize_field_name(self, name):
        normal = re.sub(r'\W', '_', name).lower()
        normal = re.sub(r'_{2,}', '_', normal)
        if normal[0] == '_':
            normal = normal[1:] 
        return normal

    def close(self):
        self.connection = None
        self.sql = None
        self.fields = None
        self.raw_fields = None

    def execute(self, sql, params=None):
        #print "EXECUTE %s" % sql
        #print params
        #print "DONE EXECUTE"

        opts = {}
        new_params = []
        # split the params into options and real params
        if params:
            for i in params:
                if type(i) == dict:
                    opts.update (i)
                else:
                    new_params.extend([i])

        if not opts.has_key ('start_at'):
            opts['start_at'] = 0
        if not opts.has_key ('max_results'):
            opts['max_results'] = -1
        if not opts.has_key ('count_only'):
            opts['count_only'] = False
        self.opts = opts
        self.sql = self.get_sql_qs(sql)
        self.sql = self.sql % tuple(self.escape_type(p) for p in new_params)

        # Only include the requested fields
        if opts.has_key('fields'):
            # The fields have the name of the 'clause' used, but we need to use
            # the 'id' to tell the API what fields to expand
            fields_s = opts['fields']
            fields = fields_s.split(',')
            # keep track of the raw fields
            self.raw_fields = self.get_raw_fields(fields)
            # replace the fields with the corresponding 'id'
            self.fields = self.get_fields(fields)
        # Check if we need to update
        if opts.has_key('update_fields'):
            rows = self.fetchmany()
            self.update(rows, opts['update_fields'])

    def fetchone(self):
        return self.fetchmany ()

    def fetchmany(self, size=None):
        max_results = self.opts['max_results']
        if size is None:
            size = self.arraysize

        if max_results < 0:
            max_results = self.arraysize
        elif max_results == 0:
            return list()
        elif max_results > self.arraysize:
            max_results = self.arraysize

        if self.opts['count_only']:
            max_results = 1

        if size < max_results:
            max_results = size

        get_opts = ''
        if self.fields and self.get_fields_qs:
            fields_qs = self.get_fields_qs(self.fields)
            if fields_qs:
                get_opts += fields_qs
                get_opts += '&'
        get_opts += self.sql
        return self.request(get_opts, max_results)

    def get_raw_fields(self, fields=None):
        content = self.request_fields(fields)
        ret = []

        if fields:
            for f in fields:
                for c in content:
                    if c.has_key('schema'):
                        if c.has_key('clauseNames') and c['clauseNames'] and c['clauseNames'][0] == f:
                            ret.append(c)
        else:
            ret = content
        return ret
 
    def get_fields(self, fields=None):
        """ 
        Method to return the FieldInfo and the field API id's
        If use_id = False, then the returned field names are the ones
        used for the query clause. If it is True, the field names
        used on the expand/select are used. The fields parameter is
        used to filter the result
        """ 
        ret = []
        if not self.raw_fields:
            self.raw_fields = self.get_raw_fields(fields)
        for f in self.raw_fields:
            if f.has_key('schema') and f.has_key('clauseNames') and f['clauseNames']:
                schema = f['schema']['type']
                array_type = None
                if schema == 'any' and f['schema']['custom']:
                    schema = f['schema']['custom']
                if schema == 'array':
                    array_type = f['schema'].get('custom', f['schema']['items'])
                editable = True
                if hasattr(self.db.introspection, 'columns_read_only'):
                    editable = f['name'] not in self.db.introspection.columns_read_only
                choices = f.get('choices', None)
 
                ret.append(FieldInfo(f['clauseNames'][0], schema, None, None, None, None, True, None, self.__normalize_field_name(f['name']), f['id'], array_type, editable, choices))
        return ret

    def escape_type(self, value):
        if type(value) == unicode or type(value) == str:
            return "'%s'" % value
        else:
            return value

    def get_sql_qs(self, sql):
        raise NotImplementedError('missing get_sql_qs implementation')

    def request(self, get_opts):
        raise NotImplementedError('missing request implementation')

    def update(self, rows, fields):
        raise NotImplementedError('missing update implementation')


class AtlassianDatabaseWrapper(BaseDatabaseWrapper):
    # Hook for sibling classes
    client_class = BaseDatabaseClient
    creation_class = AtlassianDatabaseCreation
    features_class = AtlassianDatabaseFeatures
    validation_class = AtlassianDatabaseValidation
    connection_class = AtlassianDatabaseConnection
    introspection_class = AtlassianDatabaseIntrospection
    convertion_class = AtlassianDatabaseConvertion
    charset = 'utf-8'

    Database = AtlassianDatabase

    pattern_esc = r"REPLACE(REPLACE(REPLACE({}, '\', '\\'), '%%', '\%%'), '_', '\_')"
    pattern_ops = {
        'contains': r"LIKE '%%' || {} || '%%' ESCAPE '\'",
        'icontains': r"LIKE '%%' || UPPER({}) || '%%' ESCAPE '\'",
        'startswith': r"LIKE {} || '%%' ESCAPE '\'",
        'istartswith': r"LIKE UPPER({}) || '%%' ESCAPE '\'",
        'endswith': r"LIKE '%%' || {} ESCAPE '\'",
        'iendswith': r"LIKE '%%' || UPPER({}) ESCAPE '\'",
    }
    operators = {
        'exact': '= %s',
        'iexact': 'LIKE %s',
        'contains': 'LIKE BINARY %s',
        'icontains': 'LIKE %s',
        'regex': 'REGEXP BINARY %s',
        'iregex': 'REGEXP %s',
        'gt': '> %s',
        'gte': '>= %s',
        'lt': '< %s',
        'lte': '<= %s',
        'startswith': 'LIKE BINARY %s',
        'endswith': 'LIKE BINARY %s',
        'istartswith': 'LIKE %s',
        'iendswith': 'LIKE %s',
    }

    def __init__(self, *args, **kwargs):
        super(AtlassianDatabaseWrapper, self).__init__(*args, **kwargs)
        self.autocommit = True
        self.convertion = self.convertion_class(self)

    def get_connection_params(self):
        """Compute appropriate parameters for establishing a new connection."""
        if not self.connection_class:
            raise NotImplementedError('missing connection class attribute')
        
        return self.connection_class(self.settings_dict['NAME'],
                self.settings_dict['USER'], self.settings_dict['PASSWORD'],
                self.settings_dict['SECURITY'])

    def get_new_connection(self, conn_params):
        return conn_params

    def create_cursor(self, name=None):
        if not self.cursor_class:
            raise NotImplementedError('missing cursor class attribute')
        return self.cursor_class (self.connection, self)

    def close(self):
        self.connection = None

    def init_connection_state(self):
        pass

    def _set_autocommit(self, autocommit):
        pass




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import jwt
import time
import json
import urlparse
import urllib
import hashlib
import base64
import atlassian_jwt

from django.db import models

class SecurityContext(models.base.Model):
    """
    Stores the security context shared on the installation
    handshake process
    """

    shared_secret = models.CharField(max_length=512, null=False, blank=False)
    key = models.CharField(max_length=512, null=False, blank=False)
    client_key = models.CharField(max_length=512, null=False, blank=False)
    host = models.CharField(max_length=512, null=False, blank=False)


    def create_token(self, method, uri):
        token = atlassian_jwt.encode_token(method, uri, self.key, self.shared_secret)
        return token


    def __unicode__(self):
        return "%s: %s" % (self.key, self.host)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging
import atlassian_jwt
import threading

from django.utils.deprecation import MiddlewareMixin
from django.apps import apps
from django.db import models, connections
from django.db.utils import ConnectionDoesNotExist 

from django_atlassian.models.connect import SecurityContext
from django_atlassian.models.djira import create_model, populate_model

logger = logging.getLogger('django_atlassian')
lock = threading.Lock()

class DjangoAtlassianAuthenticator(atlassian_jwt.Authenticator):
    def get_shared_secret(self, client_key):
        sc = SecurityContext.objects.filter(client_key=client_key).get()
        return sc.shared_secret


class JWTAuthenticationMiddleware(MiddlewareMixin):
    def process_request(self, request):
        headers = {}
        query = ''
        if request.method == 'POST':
            headers['Authorization'] = request.META.get('HTTP_AUTHORIZATION', None)
        # Generate the query
        params = []
        for key in request.GET.iterkeys():
            params.append("%s=%s" % (key, request.GET.get(key, None)))
        query = "&".join(params)

        auth = DjangoAtlassianAuthenticator()
        uri = request.path
        if query:
            uri = "%s?%s" % (uri, query)
        try:
            client_key = auth.authenticate(request.method, uri, headers)
        except Exception as e:
            return None

        sc = SecurityContext.objects.filter(client_key=client_key).get()
        # Setup the request attributes, the security context and the model
        try:
            db = connections[client_key]
        except ConnectionDoesNotExist:
            connections.databases[client_key] = self._create_database(client_key, sc)
            db = connections[client_key]

        with lock: 
            try:
                model = apps.get_model('django_atlassian', client_key)
            except LookupError:
                logger.info("Model %s not found, creating it", str(client_key))
                model = create_model(str(client_key))

        request.atlassian_model = model
        request.atlassian_sc = sc
        request.atlassian_db = db
        return None


    def _create_database(self, name, sc):
        new_db = {}
        new_db['id'] = name
        new_db['ENGINE'] = 'django_atlassian.backends.jira'
        new_db['NAME'] = sc.host
        new_db['USER'] = ''
        new_db['PASSWORD'] = ''
        new_db['HOST'] = ''
        new_db['PORT'] = ''
        new_db['SECURITY'] = sc
        return new_db



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import requests
import json
import logging

from django_atlassian.backends.common.base import AtlassianDatabaseWrapper, AtlassianDatabaseCursor, AtlassianDatabaseOperations, AtlassianDatabaseConvertion

logger = logging.getLogger('django_atlassian.backends.confluence')

class DatabaseCursor(AtlassianDatabaseCursor):
    arraysize = 25

    def get_fields_qs(self, fields):
        new_fields = ','.join(f[9] for f in fields if f[9] is not None)
        if new_fields:
            return 'expand=%s' % new_fields

    def get_sql_qs(self, sql):
        # For empty queries add a fake 'created' filter,
        # Confluence does not accept empty cql parameter
        query = sql.split('=')
        if not query[1]:
            query[1] = "created >= '1970/01/01'"
        return '='.join(query)


    def request_fields(self, fields=None):
        content = [
            {
                'id': None,
                'clauseNames': ['id',],
                'schema': { 'type': 'string', },
                'returnName': 'id',
            },
            {
                'id': 'history',
                'clauseNames': ['creator',],
                'schema': { 'type': 'string', },
                'returnName': 'history.createdBy.username',
            },
            {
                'id': 'history',
                'clauseNames': ['created',],
                'schema': { 'type': 'datetime', },
                'returnName': 'history.createdDate'
            },
            {
                'id': None,
                'clauseNames': ['type',],
                'schema': { 'type': 'string', },
                'returnName': 'type',
            },
            {
                'id': None,
                'clauseNames': ['title',],
                'schema': { 'type': 'string', },
                'returnName': 'title',
            },
            {
                'id': 'space',
                'clauseNames': ['space',],
                'schema': { 'type': 'string', },
                'returnName': 'space.key',
            },
        ]
        return content

    def request(self, get_opts, max_results):
        if self.opts['count_only']:
            uri_pattern = '%(uri)s/wiki/rest/api/search?%(get_opts)s&start=%(start_at)s&limit=%(max_results)s'
        else:
            uri_pattern = '%(uri)s/wiki/rest/api/content/search?%(get_opts)s&start=%(start_at)s&limit=%(max_results)s'

        uri = uri_pattern % {
            'uri': self.connection.uri,
            'get_opts': get_opts,
            'start_at': self.opts['start_at'],
            'max_results': max_results,
        }

        logger.error('uri=%s', uri)
        request = requests.get(uri, auth=(self.connection.user, self.connection.password))
        if not request:
          if self.opts['count_only']:
              return tuple([0])
          else:
              return list()

        content = json.loads(request.content)
        issues = content['results']
        num_issues = len(issues)
        if not num_issues:
            if self.opts['count_only']:
                return tuple([content['totalSize']])
            else:
                return list()

        self.opts['start_at'] += num_issues
        if (self.opts['max_results'] > 0):
            self.opts['max_results'] -= num_issues
        if self.opts['count_only']:
            return tuple([content['totalSize']])
        else:
            return tuple(self.create_row(r) for r in issues)

    def create_row(self, row):
        ret = []
        for f,r in zip(self.fields, self.raw_fields):
            obj = self.db.convertion.extract(row, f, r)
            if obj:
                ret.append(self.db.convertion.from_native(obj, f))
            else:
                ret.append(obj)
        return ret


class DatabaseConvertion(AtlassianDatabaseConvertion):
    def extract(self, data, field, raw_field):
        if raw_field.has_key('returnName'):
            return_name = raw_field['returnName']
            value = row
            for rn in return_name.split('.'):
                value = value[rn]
            return value
        elif row.has_key(field[9]):
            return row[field[9]]
        else:
            return None

class DatabaseOperations(AtlassianDatabaseOperations):
    compiler_module = "django_atlassian.backends.confluence.compiler"


class DatabaseWrapper(AtlassianDatabaseWrapper):
    vendor = 'confluence'
    cursor_class = DatabaseCursor
    ops_class = DatabaseOperations
    convertion_class = DatabaseConvertion



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf.urls import url

import views

urlpatterns = [
    url(r'^installed/$', views.installed, name='django-atlassian-installed'),
]




# -*- coding: utf-8 -*-

class Addon(object):
    def __init__(self, *args, **kwargs):
        pass

    def register(self, host, username, password, url):
        # rest/plugins/1.0/
        # reqObject.headers = {'content-type': 'application/vnd.atl.plugins.remote.install+json'}; 
        #            reqObject.body = JSON.stringify({pluginUri: descriptorUrl});
        #            reqObject.jar = false;
        #            request.post(reqObject,

    
    def unregister(self, host, username, password, url):



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import json
import logging
import requests

from django.utils.dateparse import parse_datetime
from django.db.backends.base.introspection import BaseDatabaseIntrospection, TableInfo, FieldInfo

from django_atlassian.backends.common.base import AtlassianDatabaseWrapper, AtlassianDatabaseCursor, AtlassianDatabaseOperations, AtlassianDatabaseConvertion

logger = logging.getLogger('django_atlassian.backends.jira')

class DatabaseIntrospection(BaseDatabaseIntrospection):
    data_types_reverse = {
        'datetime': 'DateField',
        'number': 'FloatField',
        'string': 'TextField',
        'array': 'django_atlassian.models.fields.ArrayField',
        'issuetype': 'TextField',
        'user': 'TextField',
        'project': 'TextField',
        'com.pyxis.greenhopper.jira:gh-epic-status': 'TextField',
    }

    data_type_relations = [
       'com.pyxis.greenhopper.jira:gh-epic-link',
       'com.atlassian.jpo:jpo-custom-field-parent',
       # Fake custom type to handle the hidden parent field
       'com.django-atlassian:parent'
    ]

    columns_read_only = [
        'Σ Time Spent', #'aggregatetimespent',
        'Σ Remaining Estimate', # 'aggregatetimeestimate',
        'Σ Original Estimate', #'aggregatetimeoriginalestimate',
        'Σ Progress', #'aggregateprogress',
        'Attachment', #'attachment',
        'Creator', #'creator',
        'Comment', #'comment',
        'Created', #'created',
        'Description', #'description',
        'Due date', #'duedate',
        'Linked Issues', #issuelinks',
        'Last Viewed', #'lastViewed',
        'Progress', #'progress',
        'Reporter', #'reporter',
        'Resolution', #'resolution',
        'Resolved', #resolutiondate',
        'Security Level', #'security',
        'Sub-tasks', #subtasks',
        'Status', #'status',
        'Remaining Estimate', #'timeestimate',
        'Original Estimate', #'timeoriginalestimate',
        'Time Spent', #timespent',
        'Time Tracking', #timetracking',
        'Updated', #updated',
        'Affects Version/s', #'versions',
        'Watchers', #'watches,',
        'Log Work', #worklog',
        'Work Ratio', #workratio',
        # Custom ones
        '[CHART] Date of First Response',
        '[CHART] Time in Status',
        'Flagged',
        'Epic/Theme',
        'Story Points',
        'Business Value',
        'Sprint',
        'Epic Link',
        'Epic Name',
        'Epic Status',
        'Epic Color',
        'Rank',
        'Organizations',
        'Parent Link', # Conditional with Portfolio
        'Change start date',
        'Change completion date',
        'Request participants',
        'Satisfaction date',
        'Approvers',
        'Story point estimate',
        'Votes',
        'Issue color',
        # Cannot find the parent fied type when saving
        # Error returned is '"parent":"data was not an object",'
        'Parent',
    ]

    def get_table_list(self, cursor):
        return [TableInfo('issue','t')]

    def get_table_description(self, cursor, table_name):
        return cursor.get_fields()

    def get_relations(self, cursor, table_name):
        relations = {}
        content = cursor.get_raw_fields()
        for f in content:
            if f.has_key('schema'):
                if f['schema']['type'] == 'any' and \
                        f['schema']['custom'] in self.data_type_relations:
                    relations[f['clauseNames'][0]] = ('issues', 'issue')
        return relations

    def get_constraints(self, cursor, table_name):
        constraints = {}
        constraints['__primary__'] = {
            'columns': ['key'],
            'primary_key': True,
            'unique': True,
            'foreign_key': False,
            'check': False,
            'index': False
        }
        return constraints


class DatabaseCursor(AtlassianDatabaseCursor):
    arraysize = 100
    uri_search_pattern = '/rest/api/3/search?%(get_opts)s&startAt=%(start_at)s&maxResults=%(max_results)s'
    uri_edit_pattern = '/rest/api/3/issue/%(issue_id)s'
    uri_field = '/rest/api/3/field'

    def get_fields_qs(self, fields):
        return 'fields=%s' % ','.join(f[9] for f in fields)

    def get_sql_qs(self, sql):
        return sql

    def update(self, rows, fields):
        self.rowcount = 0
        body = {
            'fields': {},
            'notifyUsers': False,
        }
        # sanitize the fields
        for f in self.fields:
            if not f[11]:
                continue
            for rf_key, rf_value in fields.iteritems():
                if f[0] == rf_key:
                    body['fields'].update({f[9]: self.db.convertion.to_native(rf_value, f)})
        # update every result
        # TODO Bulk change
        for row in rows:
            uri = self.uri_edit_pattern % {
                'issue_id': row[0]
            }
            response = self.connection.put_request(uri, body)
            if response.status_code not in [requests.codes.ok, 204]:
                logger.warning('JIRA Cloud returned %d for %s', response.status_code, uri)
                continue
            self.rowcount = self.rowcount + 1

    def request(self, get_opts, max_results):
        uri = self.uri_search_pattern % {
            'get_opts': get_opts,
            'start_at': self.opts['start_at'],
            'max_results': max_results,
        }

        response = self.connection.get_request(uri)
        if response.status_code != requests.codes.ok:
            logger.warning('JIRA Cloud returned %d for %s', response.status_code, uri)
            if self.opts['count_only']:
                return tuple([0])
            else:
                return list()

        content = json.loads(response.content)
        issues = content['issues']
        num_issues = len(issues)
        if not num_issues:
            if self.opts['count_only']:
                return tuple([content['total']])
            else:
                return list()

        self.opts['start_at'] += num_issues
        # In case the number is undefined, set it to the returned value
        if (self.opts['max_results'] < 0):
            self.opts['max_results'] = content['total']
        self.opts['max_results'] -= num_issues
        if self.opts['count_only']:
            return tuple([content['total']])
        else:
            return tuple(self.create_row(r) for r in issues)

    def request_fields(self, fields=None):
        """
        Get the raw description for every field id provided
        """
        # The cursor only works for the 'search' endpoint, just call
        # the 'field' endpoint and return all the field types
        response = self.connection.get_request(self.uri_field)
        if response.status_code != requests.codes.ok:
            logger.warning('JIRA Cloud returned %d for %s', response.status_code, self.uri_field)
            return []
        content = json.loads(response.content)
        # Overwrite some fields
        for c in content:
            if c['name'] == 'Epic Status':
                c['schema']['type'] = 'string'
                c['choices'] = (('To Do', 'To Do'), ('In Progress', 'In Progress'), ('Done', 'Done'))
        # The KEY field is never returned
        c = {
            "id": "key",
            "key": "key",
            "name": "Key",
            "custom": False,
            "orderable": True,
            "navigable": True,
            "searchable": True,
            "clauseNames": [
                "key",
            ],
            "schema": {
                "type": "string",
            }
        }
        content.append(c)
        # The parent field is never returned
        c = {
            "id": "parent",
            "key": "parent",
            "name": "Parent",
            "custom": True,
            "orderable": True,
            "navigable": True,
            "searchable": True,
            "clauseNames": [
                "parent",
            ],
            "schema": {
                "type": "any",
                "custom": "com.django-atlassian:parent"
            }
        }
        content.append(c)
        return content

    def create_row(self, row):
        ret = []
        for f in self.fields:
            obj = self.db.convertion.extract(row, f, None)
            if obj:
                ret.append(self.db.convertion.from_native(obj, f))
            else:
                ret.append(obj)
        return ret


class DatabaseConvertion(AtlassianDatabaseConvertion):
    def extract(self, data, field, raw_field):
        if data.has_key(field[9]):
            return data[field[9]]
        elif data['fields'].has_key(field[9]):
            return data['fields'][field[9]]
        else:
            logger.error("Field with id %s and name %s not found", field[8], field[9])
            if field[10] == 'array':
                return []
            else:
                return None

    def insert(self, data, field, raw_field):
        return {field[9]: None}

    def from_native(self, data, field):
        if field[1] == 'project' and data is not None:
            return data['key']
        # Handle the parent
        elif field[1] == 'com.django-atlassian:parent':
            if data:
                return data['key']
            else:
                return None
        # Handle the parent link
        elif field[1] == 'com.atlassian.jpo:jpo-custom-field-parent':
            if data.has_key('data'):
                return data['data']['key']
            else:
                return None
        elif field[1] == 'array' and data is not None:
            if field[10] == 'component':
                return [item['name'] for item in data]
            elif field[10] == 'string':
                return data
            elif field[10] == 'issuelinks':
                return [item['key'] for item in data]
            # Handle Jira Software special array
            elif field[10] == 'com.pyxis.greenhopper.jira:gh-sprint':
                # TODO parse the string properly in the form:
                # u'com.atlassian.greenhopper.service.sprint.Sprint@404966[id=10,rapidViewId=1,state=CLOSED,name=Sprint 6,goal=<null>,startDate=2015-05-20T12:29:00.000Z,endDate=2015-06-02T12:29:00.000Z,completeDate=2015-07-17T12:34:35.717Z,sequence=10]'
                return []
            else:
                return data
        return super(DatabaseConvertion, self).from_native(data, field)


    def to_native(self, data, field):
        if field[1] == 'project' and data is not None:
            return {'key': data}
        # Handle the parent
        elif field[1] == 'com.django-atlassian:parent':
            if data:
                return {'key': data}
            else:
                return None
        # Handle the parent link
        elif field[1] == 'com.atlassian.jpo:jpo-custom-field-parent':
            if data:
                return {'data': {'key': data}}
            else:
                return None
        elif field[1] == 'array':
            if not data:
                return []
            else:
                if field[10] == 'component':
                    return [{'name': item} for item in data]
                elif field[10] == 'string':
                    return data
                elif field[10] == 'issuelinks':
                    return [{'key': item} for item in data]
                else:
                    return []
        return super(DatabaseConvertion, self).to_native(data, field)


class DatabaseOperations(AtlassianDatabaseOperations):
    compiler_module = "django_atlassian.backends.jira.compiler"


class DatabaseWrapper(AtlassianDatabaseWrapper):
    vendor = 'jira'
    cursor_class = DatabaseCursor
    introspection_class = DatabaseIntrospection
    ops_class = DatabaseOperations
    convertion_class = DatabaseConvertion



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import json
import jwt

from django.http import HttpResponse, HttpResponseBadRequest
from django.utils.datastructures import MultiValueDictKeyError
from django.views.decorators.csrf import csrf_exempt

from models.connect import SecurityContext

@csrf_exempt
def installed(request):
    """
    Main view to handle the signal of the cloud instance when the addon
    has been installed
    """
    try:
        post = json.loads(request.body)
        key = post['key']
        shared_secret = post['sharedSecret']
        client_key = post['clientKey']
        host = post['baseUrl']
    except MultiValueDictKeyError:
        return HttpResponseBadRequest()

    # Store the security context
    # https://developer.atlassian.com/cloud/jira/platform/authentication-for-apps/
    sc = SecurityContext.objects.filter(key=key, host=host).first()
    if sc:
        update = False
        # Confirm that the shared key is the same, otherwise update it
        if sc.shared_secret != shared_secret:
            sc.shared_secret = shared_secret
            update = True
        if sc.client_key != client_key:
            sc.client_key = client_key
            update = True
        if update:
            sc.save()
    else:
        # Create a new entry on our database of connections
        sc = SecurityContext()
        sc.key = key
        sc.host = host
        sc.shared_secret = shared_secret
        sc.client_key = client_key
        sc.save()

    return HttpResponse(status=204)



import os
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
        README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-atlassian',
    version='0.1.4',
    packages=find_packages(),
    include_package_data=True,
    license='MIT License',
    description='Django app for interact with atlassian libraries such as JIRA and Confluence.',
    long_description=README,
    url='https://www.fluendo.com/',
    author='Fluendo',
    author_email='web-dev@fluendo.com',
    install_requires=[
        "Django >= 1.11",
        "PyJWT >= 1.6.4",
        "atlassian-jwt >= 1.8.1",
        "requests >= 2.18.4",
        "requests-jwt==0.5.3",
        "jira==2.0.0"],
    classifiers=[
            'Environment :: Web Environment',
            'Framework :: Django',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Topic :: Internet :: WWW/HTTP',
            'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        ],
)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import logging
import importlib
import json
import requests
import re
import collections
import os
import copy

from django.db import models, router, connections
from django.conf import settings
from django.contrib.auth.models import User
from django.db.models.signals import class_prepared
from django.db.models.manager import Manager, EmptyManager

from django_atlassian.models.connect import SecurityContext
from django_atlassian.backends.jira.base import DatabaseConvertion
from django_atlassian.models.fields import ArrayField

from jira import JIRA
from jira.resources import Issue as JiraIssue
from jira.resilientsession import ResilientSession

logger = logging.getLogger('django_atlassian')

class IssueManager(Manager):

    def create_from_json(self, json):
        """
        Some modules notify with an Issue. In order to manage it in a django way
        we need to convert the json received into a real Issue instance object
        relative to the database the manager belongs to
        """
        convert = DatabaseConvertion(None)
        args = {}
        for x in self.model.AtlassianMeta.description:
            field = self.model._meta.get_field(x[8])
            value = convert.extract(json, x, None)
            if value:
                value = convert.from_native(value, x)
            if type(field) == models.ForeignKey and value:
                args[x[8]] = self.get(key=value)
            else:
                args[x[8]] = value
        return self.model(**args)


class JiraManagerMixin(JIRA):
    def __init__(self, *args, **kwargs):
        try:
            db_alias = router.db_for_read(self.model)
            db_settings = connections.databases[db_alias]
            if db_settings['USER'] and db_settings['PASSWORD']:
                super(JiraManagerMixin, self).__init__(
                        server=db_settings['NAME'],
                        basic_auth=(
                            db_settings['USER'],
                            db_settings['PASSWORD']
                        )
                )
            elif db_settings['SECURITY']:
                jwt = {
                    'secret': db_settings['SECURITY'].shared_secret,
                    'payload': {'iss': db_settings['SECURITY'].key},
                }
                super(JiraManagerMixin, self).__init__(
                    server=db_settings['NAME'],
                    jwt=jwt
                )
        except Exception as err:
            logger.error(err)


class JiraManager(EmptyManager, JiraManagerMixin):

    def __init__(self, *args, **kwargs):
        super(JiraManager, self).__init__(None)

    def __setattr__(self, attrname, val):
        if attrname == 'model':
            if val:
                super(JiraManager, self).__init__(None)
        super(JiraManager, self).__setattr__(attrname, val)


class IssueLinkList(collections.MutableSequence):
    """
    IssueLink abstraction
    """
    uri_get = '/rest/api/3/issue/%(key)s?fields=issuelinks'
    uri_create = '/rest/api/3/issueLink'
    uri_delete = '/rest/api/3/issueLink/%(link_id)s'

    def __init__(self, model, db, link_type, inward):
        self.model = model
        self.db = db
        self.link_type = link_type
        self.inward = inward

    def __len__(self):
        uri = self.uri_get % {
            'key': self.model.key,
        }
        response = self.db.connection.get_request(uri)
        response.raise_for_status()
        content = json.loads(response.content)
        count = 0
        for x in content['fields']['issuelinks']:
            if x['type']['id'] == self.link_type['id']:
                if self.inward and x.has_key('inwardIssue'):
                    count = count + 1
                elif not self.inward and x.has_key('outwardIssue'):
                    count = count + 1
        return count

    def __delitem__(self, index):
        link_id, link = self.__getlink__(index)
        uri = self.uri_delete % {
            'link_id': link_id
        }
        response = self.db.connection.delete_request(uri)
        response.raise_for_status()

    def __setitem__(self, index, value):
        raise NotImplementedError

    def insert(self, index, value):
        if not isinstance(value, self.model._meta.model):
            raise TypeError('The value must be an Issue')

        if index != len(self):
            raise IndexError
        body = {
            'type': { 'name': self.link_type['name'] },
        }
        if self.inward:
            body['inwardIssue'] = {'key': value.key}
            body['outwardIssue'] = {'key': self.model.key}
        else:
            body['inwardIssue'] = {'key': self.model.key}
            body['outwardIssue'] = {'key': value.key}

        response = self.db.connection.post_request(self.uri_create, body)
        response.raise_for_status()

    def __getitem__(self, index):
        link_id, link = self.__getlink__(index)
        key = link['key']
        return self.model._meta.model.objects.get(key=key)

    def __getlink__(self, index):
        uri = self.uri_get % {
            'key': self.model.key,
        }
        response = self.db.connection.get_request(uri)
        response.raise_for_status()

        content = json.loads(response.content)
        count = 0
        link = None
        link_id = 0
        for x in content['fields']['issuelinks']:
            if x['type']['id'] == self.link_type['id']:
                if self.inward and x.has_key('inwardIssue'):
                    if count == index:
                        link = x['inwardIssue']
                        link_id = x['id']
                        break
                    count = count + 1
                elif not self.inward and x.has_key('outwardIssue'):
                    if count == index:
                        link = x['outwardIssue']
                        link_id = x['id']
                        break
                    count = count + 1
        if not link:
            raise IndexError("list index out of range")
        return link_id, link


class IssueLinks(object):
    """
    IssueLinks abstraction
    """

    db = None
    key = None
    uri_get_all = '/rest/api/3/issueLinkType'
    uri_get = '/rest/api/3/issue/%(key)s?fields=issuelinks'

    def __init__(self, model):
        db_alias = router.db_for_read(model._meta.model)
        db_settings = connections.databases[db_alias]
        if db_settings['ENGINE'] != 'django_atlassian.backends.jira':
            return

        self.db = connections[db_alias]
        self.model = model

    def __normalize(self, name):
        normal = re.sub(r'\W', '_', name).lower()
        normal = re.sub(r'_{2,}', '_', normal)
        return normal

    def __getattr__(self, name):
        # Get every link type
        response = self.db.connection.get_request(self.uri_get_all)
        if response.status_code != requests.codes.ok:
            raise AttributeError("'IssueLinks' has no attribute %s" % name)
        
        content = json.loads(response.content)
        link_type = None
        inward = False
        for x in content['issueLinkTypes']:
            if name == self.__normalize(x['inward']):
                inward = True
                link_type = x
                break
            elif name == self.__normalize(x['outward']):
                inward = False
                link_type = x
                break

        if not link_type:
            raise AttributeError("'IssueLinks' has no attribute %s" % name)

        return IssueLinkList(self.model, self.db, link_type, inward)


class Attachment(object):
    """
    Attachment class for jira issue
    """
    def __init__(self, obj, db):
        self.db = db
        for key, value in obj.iteritems():
            if isinstance(value, dict):
               self.__init__(value, db)
            else:
                setattr(self, key, value)

    def __getattr__(self, attr):
        return "'Attachment' attribute {} not defined".format(attr)

    def delete(self):
        uri = '/rest/api/3/attachment/{}'.format(self.id)
        try:
            response = self.db.connection.delete_request(uri)
            return response.status_code
        except Exception as e:
            raise e

class JiraIssueModel(JiraIssue):
    def __init__(self, *args, **kwargs):
        options = copy.copy(JIRA.DEFAULT_OPTIONS)
        session = ResilientSession()
        jwt = None
        if self.AtlassianMeta.db:
            # dynamic models via jira-addon
            db = connections.databases[self.AtlassianMeta.db]
            options['server'] = db['NAME']
            sc = db['SECURITY']
            jwt = {
                'secret': sc.shared_secret,
                'payload': {'iss': sc.key},
            }
        else:
            # static models 
            db = self.get_db().connection
            options['server'] = db.uri
            session.auth = (db.user, db.password)

        # Jira model init self.key = None, keep a copy & restore
        self.jira_key = self.key
        super(JiraIssueModel, self).__init__(options, session)
        self.key = self.jira_key
        if jwt:
            self._session = self.__class__.jira._session


class Issue(models.base.Model, JiraIssueModel):
    """
    Base class for all JIRA Issue models.
    """

    # The only mandatory field we add here is the KEY. For half-dynamic models
    # it is required to have a primary key otherwise django will create one for
    # us.
    key = models.CharField(max_length=255, primary_key=True, unique=True)
    objects = IssueManager()
    jira = JiraManager()

    def __init__(self, *args, **kwargs):
        super(Issue, self).__init__(*args, **kwargs)
        self.find(self.jira_key)

    def __getattr__(self, name):
        if name == 'links':
            self.links = IssueLinks(self)
            return self.links
        raise AttributeError("'Issue' has no attribute %s" % name)

    def get_db(self):
        db_alias = router.db_for_read(self._meta.model)
        db_settings = connections.databases[db_alias]
        if db_settings['ENGINE'] != 'django_atlassian.backends.jira':
            return None
        return connections[db_alias]

    def get_property(self, prop_name):
        # Create a connection
        # Call curl -X GET https://jira-instance1.net/rest/api/3/issue/ENPR-4/properties/{propertyKey}
        pass

    def set_property(self, prop_name, value):
        # Create a connection
        # Call curl -X PUT -H "Content-type: application/json" https://jira-instance1.net/rest/api/3/issue/`ENPR-4`/properties/{propertyKey} -d '{"content":"Test if works on Jira Cloud", "completed" : 1}'
        pass

    def is_parent(self):
        """
        Return True if the issue is a parent issue, False otherwise
        """
        if self.epic_linked.exists():
            return True
        elif self.sub_tasks:
            return True
        else:
            return False

    def has_parent(self):
        """ 
        Return True if the issue has a parent, False otherwise
        """
        if self.epic_link_id is not None:
            return True
        if self.parent_link_id is not None:
            return True
        if self.parent_id is not None:
            return True
        return False 

    def get_parent(self):
        """ 
        Get the parent issue depending on the issue type
        """
        if self.epic_link_id is not None:
            return self.epic_link
        if self.parent_link_id is not None:
            return self.parent_link
        if self.parent_id is not None:
            return self.parent
        return None

    def get_children(self):
        """
        Get the children issues depending on the isue type
        """
        if self.epic_linked.exists():
            return self.epic_linked.all()
        elif self.sub_tasks:
            return self._meta.model.objects.filter(key__in=self.sub_tasks)
        else:
            return []

    def get_changelog(self):
        """
        Get the Issue's changelog
        """
        uri = "/rest/api/3/issue/%(issue)s/changelog" % { 'issue': self.key }
        response = self.get_db().connection.get_request(uri)
        response.raise_for_status()
        content = json.loads(response.content)
        return content

    def get_statuses(self):
        """
        Get the available statuses on the system
        """
        uri = "/rest/api/3/status"
        response = self.get_db().connection.get_request(uri)
        response.raise_for_status()
        content = json.loads(response.content)
        return content
    
    def get_attachment(self):
        """
        Get an url list of attached files
        """
        response = self.attachment
        content = [  Attachment(file, self.get_db()) for file in response]
        return content

    def add_attachment(self, attachment, filename=None):
        """
        Attach a new file
        :param attachment: a file like object
        :param filename: a file name
        """
        uri = "/rest/api/3/issue/{}/attachments".format(self.key)
        if isinstance(attachment, str):
            attachment = open(attachment, "rb") 

        if not filename:
            filename = os.path.basename(attachment.name)

        try:
            response = self.get_db().connection.post_request(uri, body={}, header={'content-type': None, 'X-Atlassian-Token': 'nocheck'},
                                                  fil={'file': (filename, attachment, 'application/octet-stream')})
            return response
        except Exception as e:
            raise e

    def delete_attachment(self, attachment):
        """
        Delete attachment file
        :param attachment: it is an Attachment object
        """
        uri = '/rest/api/3/attachment/{}'.format(attachment.id)

        try:
            response = self.get_db().connection.delete_request(uri)
            return response.status_code
        except Exception as e:
            raise e 

    def __unicode__(self):
        return str(self.key)

    class Meta:
        abstract = True


class AtlassianMeta:
    """
    Base class for all JIRA related Meta.
    """

    def __init__(self):
        # The database name this model refers to. Even if this breaks django
        # purpose of resuable models being independent of the database backend,
        # for JIRA there's 1:1 relation between the database (connection) and the model
        self.db = None
        # The set of FieldInfo as returned by the introspection
        # get_table_description(). Given that several REST API methods return a
        # full issue in JSON, we can parse it directly without the need of another
        # round trip to the server
        self.description = []


def create_model(name):
    # Create the module dynamically
    class Meta:
        pass
    setattr(Meta, 'app_label', 'django_atlassian')
    setattr(Meta, 'managed', False)

    am = AtlassianMeta()
    am.db = name

    # Set up a dictionary to simulate declarations within a class
    attrs = {
        '__module__': create_model.__module__, 
        'Meta': Meta,
        'AtlassianMeta': am
    }
    logger.info("Creating model %s", name)
    model = type(name, (Issue,), attrs)
    return model


def populate_model(db, model):
    # Double check the model does have an AtlassianMeta
    am = getattr(model, 'AtlassianMeta', None)
    if not am:
        am = AtlassianMeta()
        setattr(model, 'AtlassianMeta', am)
    # Check if the description is already populated
    if am.description:
        return
        
    logger.info("Populating model %s", model)
    # Create a cursor to inspect the database for the fields
    with db.cursor() as cursor:
        try:
            relations = db.introspection.get_relations(cursor, 'issue')
        except NotImplementedError:
            relations = {}
        try:
            constraints = db.introspection.get_constraints(cursor, 'issue')
        except NotImplementedError:
            constraints = {}
        primary_key_column = db.introspection.get_primary_key_column(cursor, 'issue')
        unique_columns = [
            c['columns'][0] for c in constraints.values()
            if c['unique'] and len(c['columns']) == 1
        ]
        table_description = db.introspection.get_table_description(cursor, 'issue')
        for row in table_description:
            extra_params = {}
            column_name = row[0]
            field_name = row[8]
            choices = row[12]
            is_relation = column_name in relations

            # Use the correct column name
            extra_params['db_column'] = column_name
            # Skip the primary key, we already have one
            if column_name == primary_key_column:
                am.description.append(row)
                continue
            # Add unique, if necessary.
            if column_name in unique_columns:
                extra_params['unique'] = True

            # Add choices if needed
            if choices:
                extra_params['choices'] = choices

            if is_relation:
                field_type = 'ForeignKey'
                rel_to = (
                    "self" if relations[column_name][1] == 'issue'
                    else None
                )
            else:
               try:
                   field_type = db.introspection.get_field_type(row[1], row)
               except KeyError:
                   field_type = None

            if field_type:
                field_module = 'django.db.models' if '.' not in field_type else '.'.join(field_type.split('.')[:-1])
                field_class = field_type if '.' not in field_type else field_type.split('.')[-1]
                logger.info("Adding field '%s' for column '%s' for class '%s'", field_name, column_name, field_class) 
                try:
                    field_module = importlib.import_module(field_module)
                    try:
                        field_cls = getattr(field_module, field_class)
                    except AttributeError:
                        logger.error("Class '%s' does not exist", field_class)
                        continue
                except ImportError:
                    logger.error("Module '%s' does not exist", field_module)
                    continue

                if field_type == 'ForeignKey' and rel_to:
                    field = field_cls(rel_to, related_name='%sed' % field_name, **extra_params)
                else:
                    field = field_cls(**extra_params)

                if field:
                    am.description.append(row)
                    field.contribute_to_class(model, field_name)
                else:
                    logger.warning("Field '%s' can not be added", field_name)


def add_fields(sender, **kwargs):
    """
    Will add the corresponding fields for the database this
    model belongs to
    """
    db_alias = router.db_for_read(sender)
    db = connections.databases[db_alias]
    if db['ENGINE'] != 'django_atlassian.backends.jira':
        return

    connection = connections[db_alias]
    logger.info("Class %s prepared, populating", sender)
    populate_model(connection, sender)

 
class_prepared.connect(add_fields)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.contrib.auth import get_user_model
from django.db.models.sql import compiler
from django.db.models.expressions import Col
from django.db.models.aggregates import Count
from django.db.models.fields.related_lookups import RelatedExact

class SQLCompiler(compiler.SQLCompiler):

    def compile(self, node, select_format=False):
        # A column must not have a table.column form
        if type(node) == Col:
            return '%s' % node.target.column, []
        # Skip the Count
        elif type(node) == Count:
            return '', [{'count_only': True}]
        return super(SQLCompiler, self).compile(node, select_format)

    def as_sql(self, with_limits=True, with_col_aliases=False):
        """
        Creates the SQL for this query. Returns the SQL string and list of
        parameters.

        If 'with_limits' is False, any limit/offset information is not included
        in the query.
        """
        refcounts_before = self.query.alias_refcount.copy()
        try:
            extra_select, order_by, group_by = self.pre_sql_setup()

            # This must come after 'select', 'ordering', and 'distinct' -- see
            # docstring of get_from_clause() for details.
            from_, f_params = self.get_from_clause()

            where, w_params = self.compile(self.where) if self.where is not None else ("", [])
            having, h_params = self.compile(self.having) if self.having is not None else ("", [])


            # [J,C]QL doesn't support any combinator
            if self.query.combinator:
                raise DatabaseError('{} not supported on this database backend.'.format(combinator))
            params = []

            out_cols = []
            for _, (s_sql, s_params), alias in self.select + extra_select:
                params.extend(s_params)
                out_cols.append(s_sql)

            if out_cols:
                params.extend([{'fields': ','.join(out_cols)}])

            # JQL doesn't support a select for update
            if self.query.select_for_update:
                raise TransactionManagementError('select_for_update cannot be used outside of a transaction.')

            sql = ''
            if where:
                sql += where + ' '
                params.extend(w_params)

            # JQL doesn't support group by
            if group_by:
                raise DatabaseError('GROUP BY is not supported on this database backend.')

            if having:
                raise DatabaseError('HAVING is not supported on this database backend.')

            if order_by:
                ordering = []
                for _, (o_sql, o_params, _) in order_by:
                    ordering.append(o_sql)
                    params.extend(o_params)
                sql += 'ORDER BY %s' % ', '.join(ordering)

            if with_limits:
                if self.query.high_mark is not None:
                    params.extend([{'max_results': self.query.high_mark - self.query.low_mark}])
                if self.query.low_mark:
                    params.extend([{'start_at': self.query.low_mark}])
            return "%s=%s" % (self.query_param, sql), tuple(params)
        finally:
            # Finally do cleanup - get rid of the joins we created above.
            self.query.reset_refcounts(refcounts_before)
